@@ -6,14 +6,19 @@
 
 {-# LANGUAGE CPP, Rank2Types, ScopedTypeVariables, FlexibleContexts, FlexibleInstances, GeneralizedNewtypeDeriving #-}
 {-# LANGUAGE ExistentialQuantification #-}
+{- HLINT ignore "Use camelCase" -}
 
 module Main where
 
-import Prelude hiding (foldl, foldr, gcd, length, null, reverse, span, splitAt, takeWhile)
+import Prelude (Bool(..), Ordering, Int, Integer, Double, Float, Char, String,
+                Maybe(..), Either(..), Eq, Show, (.), ($), (*), (==), (/=),
+                (&&), (||), (++), (>>=), fmap, maybe, either, map, all, not,
+                undefined, const, flip, succ, uncurry, min, id, replicate,
+                minBound, maxBound, otherwise, fst, snd, concatMap, mappend, div)
 
 import Test.Tasty (defaultMain, testGroup)
 import Test.Tasty.QuickCheck (Arbitrary, CoArbitrary, Property, Gen,
-                              arbitrary, coarbitrary, property, label, forAll, mapSize, testProperty, variant, whenFail, (.&&.))
+                              arbitrary, coarbitrary, property, label, forAll, mapSize, testProperty, variant, whenFail, (.&&.), (===))
 import Test.QuickCheck.Instances ()
 
 import Control.Applicative (Applicative(..), liftA2)
@@ -53,6 +58,7 @@ import Data.Monoid.Instances.Concat (Concat)
 import qualified Data.Monoid.Instances.Concat as Concat
 import Data.Monoid.Instances.Measured (Measured)
 import qualified Data.Monoid.Instances.Measured as Measured
+import qualified Data.Monoid.Instances.PrefixMemory as PrefixMemory
 import Data.Monoid.Instances.Stateful (Stateful)
 import qualified Data.Monoid.Instances.Stateful as Stateful
 import Data.Monoid.Instances.Positioned (OffsetPositioned, LinePositioned)
@@ -70,10 +76,26 @@ import Data.Semigroup.Cancellative (Commutative, Reductive,
 import Data.Monoid.Null (MonoidNull, PositiveMonoid, null)
 import Data.Monoid.Factorial (FactorialMonoid,
                               splitPrimePrefix, splitPrimeSuffix, inits, tails, span, spanMaybe, split, splitAt)
-import Data.Monoid.GCD (GCDMonoid, LeftGCDMonoid, RightGCDMonoid, gcd,
-                        commonPrefix, stripCommonPrefix,
-                        commonSuffix, stripCommonSuffix)
-import Data.Monoid.Monus (OverlappingGCDMonoid, Monus, (<\>), stripPrefixOverlap, stripSuffixOverlap)
+import Data.Monoid.GCD
+    ( GCDMonoid
+    , LeftGCDMonoid
+    , RightGCDMonoid
+    , DistributiveGCDMonoid
+    , LeftDistributiveGCDMonoid
+    , RightDistributiveGCDMonoid
+    , commonPrefix
+    , commonSuffix
+    , gcd
+    , stripCommonPrefix
+    , stripCommonSuffix
+    )
+import Data.Monoid.LCM
+    ( LCMMonoid
+    , DistributiveLCMMonoid
+    , lcm
+    )
+import Data.Monoid.Monus (OverlappingGCDMonoid, Monus,
+                          (<\>), overlap, stripOverlap, stripPrefixOverlap, stripSuffixOverlap)
 import Data.Monoid.Textual (TextualMonoid)
 import qualified Data.Monoid.Textual as Textual
 
@@ -94,7 +116,11 @@ data Test = CommutativeTest (CommutativeMonoidInstance -> Property)
           | LeftGCDTest (LeftGCDMonoidInstance -> Property)
           | RightGCDTest (RightGCDMonoidInstance -> Property)
           | GCDTest (GCDMonoidInstance -> Property)
-          | CancellativeGCDTest (CancellativeGCDMonoidInstance -> Property)
+          | DistributiveGCDTest (DistributiveGCDMonoidInstance -> Property)
+          | LeftDistributiveGCDTest (LeftDistributiveGCDMonoidInstance -> Property)
+          | RightDistributiveGCDTest (RightDistributiveGCDMonoidInstance -> Property)
+          | LCMTest (LCMMonoidInstance -> Property)
+          | DistributiveLCMTest (DistributiveLCMMonoidInstance -> Property)
 
 data CommutativeMonoidInstance = forall a. (Arbitrary a, Show a, Eq a, Commutative a, Monoid a) =>
                                  CommutativeMonoidInstance a
@@ -135,8 +161,26 @@ data RightGCDMonoidInstance = forall a. (Arbitrary a, Show a, Eq a, RightGCDMono
                               RightGCDMonoidInstance a
 data GCDMonoidInstance = forall a. (Arbitrary a, Show a, Eq a, GCDMonoid a) =>
                          GCDMonoidInstance a
-data CancellativeGCDMonoidInstance = forall a. (Arbitrary a, Show a, Eq a, Monoid a, Cancellative a, GCDMonoid a) =>
-                                     CancellativeGCDMonoidInstance a
+
+data DistributiveGCDMonoidInstance =
+    forall a. (Arbitrary a, Show a, Eq a, DistributiveGCDMonoid a)
+        => DistributiveGCDMonoidInstance a
+
+data LeftDistributiveGCDMonoidInstance =
+    forall a. (Arbitrary a, Show a, Eq a, LeftDistributiveGCDMonoid a)
+        => LeftDistributiveGCDMonoidInstance a
+
+data RightDistributiveGCDMonoidInstance =
+    forall a. (Arbitrary a, Show a, Eq a, RightDistributiveGCDMonoid a)
+        => RightDistributiveGCDMonoidInstance a
+
+data LCMMonoidInstance =
+    forall a. (Arbitrary a, Show a, Eq a, LCMMonoid a)
+        => LCMMonoidInstance a
+
+data DistributiveLCMMonoidInstance =
+    forall a. (Arbitrary a, Show a, Eq a, DistributiveLCMMonoid a)
+        => DistributiveLCMMonoidInstance a
 
 commutativeInstances :: [CommutativeMonoidInstance]
 commutativeInstances = map upcast reductiveInstances
@@ -195,7 +239,7 @@ factorialInstances = map upcast stableFactorialInstances
    where upcast (StableFactorialMonoidInstance i) = FactorialMonoidInstance i
 
 stableFactorialInstances :: [StableFactorialMonoidInstance]
-stableFactorialInstances = stable1 ++ map measure stable1 ++ map position stable1 
+stableFactorialInstances = stable1 ++ map measure stable1 ++ map prefixed stable1 ++ map position stable1 
    where stable1 = map upcast stableTextualInstances
                    ++ [StableFactorialMonoidInstance (mempty :: ByteString),
                        StableFactorialMonoidInstance (mempty :: Lazy.ByteString),
@@ -204,6 +248,7 @@ stableFactorialInstances = stable1 ++ map measure stable1 ++ map position stable
                        StableFactorialMonoidInstance (mempty :: Vector Int)]
          upcast (StableTextualMonoidInstance i) = StableFactorialMonoidInstance i
          measure (StableFactorialMonoidInstance i) = StableFactorialMonoidInstance (Measured.measure i)
+         prefixed (StableFactorialMonoidInstance i) = StableFactorialMonoidInstance (PrefixMemory.shadowed i)
          position (StableFactorialMonoidInstance (i :: a)) = 
             StableFactorialMonoidInstance (pure i :: OffsetPositioned a)
 
@@ -220,7 +265,7 @@ textualInstances = map upcast stableTextualInstances
    where upcast (StableTextualMonoidInstance i) = TextualMonoidInstance i
 
 stableTextualInstances :: [StableTextualMonoidInstance]
-stableTextualInstances = stable1 ++ map measure stable1 ++ concatMap position stable1
+stableTextualInstances = stable1 ++ map measure stable1 ++ map prefixed stable1 ++ concatMap position stable1
    where stable1 = [StableTextualMonoidInstance (mempty :: TestString),
                     StableTextualMonoidInstance (mempty :: String),
                     StableTextualMonoidInstance (mempty :: Text),
@@ -228,6 +273,7 @@ stableTextualInstances = stable1 ++ map measure stable1 ++ concatMap position st
                     StableTextualMonoidInstance (mempty :: Seq Char),
                     StableTextualMonoidInstance (mempty :: Vector Char)]
          measure (StableTextualMonoidInstance i) = StableTextualMonoidInstance (Measured.measure i)
+         prefixed (StableTextualMonoidInstance i) = StableTextualMonoidInstance (PrefixMemory.shadowed i)
          position (StableTextualMonoidInstance (i :: a)) = 
             [StableTextualMonoidInstance (pure i :: OffsetPositioned a),
              StableTextualMonoidInstance (pure i :: LinePositioned a)]
@@ -247,6 +293,7 @@ leftReductiveInstances = map upcast leftCancellativeInstances
                              LeftReductiveMonoidInstance (mempty :: LinePositioned Text),
                              LeftReductiveMonoidInstance (mempty :: OffsetPositioned Text),
                              LeftReductiveMonoidInstance (mempty :: Measured Text),
+                             LeftReductiveMonoidInstance (mempty :: PrefixMemory.Shadowed Text),
                              LeftReductiveMonoidInstance (mempty :: Stateful (Sum Integer) Text)]
    where upcast (LeftCancellativeMonoidInstance i) = LeftReductiveMonoidInstance i
 
@@ -264,6 +311,7 @@ rightReductiveInstances = map upcast rightCancellativeInstances
                               RightReductiveMonoidInstance (mempty :: LinePositioned Text),
                               RightReductiveMonoidInstance (mempty :: OffsetPositioned Text),
                               RightReductiveMonoidInstance (mempty :: Measured Text),
+                              RightReductiveMonoidInstance (mempty :: PrefixMemory.Shadowed Text),
                               RightReductiveMonoidInstance (mempty :: Stateful (Sum Integer) Text)]
    where upcast (RightCancellativeMonoidInstance i) = RightReductiveMonoidInstance i
 
@@ -289,7 +337,9 @@ overlappingGCDMonoidInstances = map upcast monusInstances
 monusInstances = [MonusInstance (mempty :: Product Natural),
                   MonusInstance (mempty :: Sum Natural),
                   MonusInstance (mempty :: Dual (Product Natural)),
+                  MonusInstance (mempty :: Maybe ()),
                   MonusInstance (mempty :: Maybe (Product Natural)),
+                  MonusInstance (mempty :: Maybe (Sum Natural)),
                   MonusInstance (mempty :: IntSet),
                   MonusInstance (mempty :: Set String)]
 
@@ -316,9 +366,7 @@ rightCancellativeInstances = map upcast cancellativeInstances
                                 RightCancellativeMonoidInstance (mempty :: Vector Int)]
    where upcast (CancellativeMonoidInstance i) = RightCancellativeMonoidInstance i
 
-cancellativeInstances = map upcast cancellativeGCDInstances
-                        ++ []
-   where upcast (CancellativeGCDMonoidInstance i) = CancellativeMonoidInstance i
+cancellativeInstances = [CancellativeMonoidInstance ()]
 
 leftGCDInstances = map upcast gcdInstances
                    ++ [LeftGCDMonoidInstance (mempty :: String),
@@ -357,17 +405,118 @@ rightGCDInstances = map upcast gcdInstances
                        RightGCDMonoidInstance (mempty :: Concat (Dual Text))]
    where upcast (GCDMonoidInstance i) = RightGCDMonoidInstance i
 
-gcdInstances = map upcast cancellativeGCDInstances
-               ++ [GCDMonoidInstance (mempty :: Product Natural),
-                   GCDMonoidInstance (mempty :: Dual (Product Natural)),
-                   GCDMonoidInstance (mempty :: IntSet),
-                   GCDMonoidInstance (mempty :: Set String)]
-   where upcast (CancellativeGCDMonoidInstance i) = GCDMonoidInstance i
+gcdInstances =
+    [ GCDMonoidInstance (mempty :: ())
+    , GCDMonoidInstance (mempty :: Product Natural)
+    , GCDMonoidInstance (mempty :: Dual (Product Natural))
+    , GCDMonoidInstance (mempty :: IntSet)
+    , GCDMonoidInstance (mempty :: Set String)
+    ]
 
-cancellativeGCDInstances = [CancellativeGCDMonoidInstance ()]
+distributiveGCDMonoidInstances :: [DistributiveGCDMonoidInstance]
+distributiveGCDMonoidInstances =
+    [ DistributiveGCDMonoidInstance (mempty :: ())
+    , DistributiveGCDMonoidInstance (mempty :: Product Natural)
+    , DistributiveGCDMonoidInstance (mempty :: Sum Natural)
+    , DistributiveGCDMonoidInstance (mempty :: IntSet)
+    , DistributiveGCDMonoidInstance (mempty :: Set ())
+    , DistributiveGCDMonoidInstance (mempty :: Set Bool)
+    , DistributiveGCDMonoidInstance (mempty :: Set Word)
+    , DistributiveGCDMonoidInstance (mempty :: Dual (Set ()))
+    , DistributiveGCDMonoidInstance (mempty :: Dual (Set Bool))
+    , DistributiveGCDMonoidInstance (mempty :: Dual (Set Word))
+    ]
+
+leftDistributiveGCDMonoidInstances :: [LeftDistributiveGCDMonoidInstance]
+leftDistributiveGCDMonoidInstances =
+    [ -- Instances for non-commutative monoids:
+      LeftDistributiveGCDMonoidInstance (mempty :: [()])
+    , LeftDistributiveGCDMonoidInstance (mempty :: [Bool])
+    , LeftDistributiveGCDMonoidInstance (mempty :: [Word])
+    , LeftDistributiveGCDMonoidInstance (mempty :: Seq ())
+    , LeftDistributiveGCDMonoidInstance (mempty :: Seq Bool)
+    , LeftDistributiveGCDMonoidInstance (mempty :: Seq Word)
+    , LeftDistributiveGCDMonoidInstance (mempty :: Vector ())
+    , LeftDistributiveGCDMonoidInstance (mempty :: Vector Bool)
+    , LeftDistributiveGCDMonoidInstance (mempty :: Vector Word)
+    , LeftDistributiveGCDMonoidInstance (mempty :: ByteString)
+    , LeftDistributiveGCDMonoidInstance (mempty :: Lazy.ByteString)
+    , LeftDistributiveGCDMonoidInstance (mempty :: Text)
+    , LeftDistributiveGCDMonoidInstance (mempty :: Lazy.Text)
+      -- Instances for commutative monoids:
+    , LeftDistributiveGCDMonoidInstance (mempty :: ())
+    , LeftDistributiveGCDMonoidInstance (mempty :: Product Natural)
+    , LeftDistributiveGCDMonoidInstance (mempty :: Sum Natural)
+    , LeftDistributiveGCDMonoidInstance (mempty :: IntSet)
+    , LeftDistributiveGCDMonoidInstance (mempty :: Set ())
+    , LeftDistributiveGCDMonoidInstance (mempty :: Set Bool)
+    , LeftDistributiveGCDMonoidInstance (mempty :: Set Word)
+      -- Instances for monoid transformers:
+    , LeftDistributiveGCDMonoidInstance (mempty :: Dual [()])
+    , LeftDistributiveGCDMonoidInstance (mempty :: Dual [Bool])
+    , LeftDistributiveGCDMonoidInstance (mempty :: Dual [Word])
+    ]
+
+rightDistributiveGCDMonoidInstances :: [RightDistributiveGCDMonoidInstance]
+rightDistributiveGCDMonoidInstances =
+    [ -- Instances for non-commutative monoids:
+      RightDistributiveGCDMonoidInstance (mempty :: [()])
+    , RightDistributiveGCDMonoidInstance (mempty :: [Bool])
+    , RightDistributiveGCDMonoidInstance (mempty :: [Word])
+    , RightDistributiveGCDMonoidInstance (mempty :: Seq ())
+    , RightDistributiveGCDMonoidInstance (mempty :: Seq Bool)
+    , RightDistributiveGCDMonoidInstance (mempty :: Seq Word)
+    , RightDistributiveGCDMonoidInstance (mempty :: Vector ())
+    , RightDistributiveGCDMonoidInstance (mempty :: Vector Bool)
+    , RightDistributiveGCDMonoidInstance (mempty :: Vector Word)
+    , RightDistributiveGCDMonoidInstance (mempty :: ByteString)
+    , RightDistributiveGCDMonoidInstance (mempty :: Lazy.ByteString)
+    , RightDistributiveGCDMonoidInstance (mempty :: Text)
+    , RightDistributiveGCDMonoidInstance (mempty :: Lazy.Text)
+      -- Instances for commutative monoids:
+    , RightDistributiveGCDMonoidInstance (mempty :: ())
+    , RightDistributiveGCDMonoidInstance (mempty :: Product Natural)
+    , RightDistributiveGCDMonoidInstance (mempty :: Sum Natural)
+    , RightDistributiveGCDMonoidInstance (mempty :: IntSet)
+    , RightDistributiveGCDMonoidInstance (mempty :: Set ())
+    , RightDistributiveGCDMonoidInstance (mempty :: Set Bool)
+    , RightDistributiveGCDMonoidInstance (mempty :: Set Word)
+      -- Instances for monoid transformers:
+    , RightDistributiveGCDMonoidInstance (mempty :: Dual [()])
+    , RightDistributiveGCDMonoidInstance (mempty :: Dual [Bool])
+    , RightDistributiveGCDMonoidInstance (mempty :: Dual [Word])
+    ]
+
+lcmInstances =
+    [LCMMonoidInstance (mempty :: Product Natural),
+     LCMMonoidInstance (mempty :: Sum Natural),
+     LCMMonoidInstance (mempty :: Dual (Product Natural)),
+     LCMMonoidInstance (mempty :: Dual (Sum Natural)),
+     LCMMonoidInstance (mempty :: IntSet),
+     LCMMonoidInstance (mempty :: (IntSet, IntSet)),
+     LCMMonoidInstance (mempty :: (IntSet, IntSet, IntSet)),
+     LCMMonoidInstance (mempty :: (IntSet, IntSet, IntSet, IntSet)),
+     -- For sets, test with a variety of different universe sizes, from small
+     -- to large:
+     LCMMonoidInstance (mempty :: Set ()),
+     LCMMonoidInstance (mempty :: Set Bool),
+     LCMMonoidInstance (mempty :: Set Ordering),
+     LCMMonoidInstance (mempty :: Set Word8)]
+
+distributiveLCMInstances =
+    [ DistributiveLCMMonoidInstance (mempty :: ())
+    , DistributiveLCMMonoidInstance (mempty :: Product Natural)
+    , DistributiveLCMMonoidInstance (mempty :: Sum Natural)
+    , DistributiveLCMMonoidInstance (mempty :: IntSet)
+    , DistributiveLCMMonoidInstance (mempty :: Set ())
+    , DistributiveLCMMonoidInstance (mempty :: Set Bool)
+    , DistributiveLCMMonoidInstance (mempty :: Set Word)
+    , DistributiveLCMMonoidInstance (mempty :: Dual (Product Natural))
+    , DistributiveLCMMonoidInstance (mempty :: Dual (Sum Natural))
+    ]
 
 main = defaultMain (testGroup "MonoidSubclasses" $ map expand tests)
-  where expand (name, test) = testProperty name (foldr1 (.&&.) $ checkInstances test)
+  where expand (name, test) = testProperty name (List.foldr1 (.&&.) $ checkInstances test)
 
 checkInstances :: Test -> [Property]
 checkInstances (CommutativeTest checkType) = (map checkType commutativeInstances)
@@ -387,7 +536,11 @@ checkInstances (MonusTest checkType) = (map checkType monusInstances)
 checkInstances (LeftGCDTest checkType) = (map checkType leftGCDInstances) 
 checkInstances (RightGCDTest checkType) = (map checkType rightGCDInstances) 
 checkInstances (GCDTest checkType) = (map checkType gcdInstances)  
-checkInstances (CancellativeGCDTest checkType) = (map checkType cancellativeGCDInstances) 
+checkInstances (DistributiveGCDTest checkType) = (map checkType distributiveGCDMonoidInstances)
+checkInstances (LeftDistributiveGCDTest checkType) = (map checkType leftDistributiveGCDMonoidInstances)
+checkInstances (RightDistributiveGCDTest checkType) = (map checkType rightDistributiveGCDMonoidInstances)
+checkInstances (LCMTest checkType) = (map checkType lcmInstances)
+checkInstances (DistributiveLCMTest checkType) = (map checkType distributiveLCMInstances)
 
 tests :: [(String, Test)]
 tests = [("CommutativeMonoid", CommutativeTest checkCommutative),
@@ -456,6 +609,12 @@ tests = [("CommutativeMonoid", CommutativeTest checkCommutative),
          ("stripSuffixOverlap 1", OverlappingGCDTest checkStripSuffixOverlap1),
          ("stripSuffixOverlap 2", OverlappingGCDTest checkStripSuffixOverlap2),
          ("stripSuffixOverlap 3", OverlappingGCDTest checkStripSuffixOverlap3),
+         ("overlap law 1", OverlappingGCDTest checkOverlapLaw1),
+         ("overlap law 2", OverlappingGCDTest checkOverlapLaw2),
+         ("overlap law 3", OverlappingGCDTest checkOverlapLaw3),
+         ("overlap idempotence", OverlappingGCDTest checkOverlap_idempotence),
+         ("overlap identity (left)", OverlappingGCDTest checkOverlap_identity_left),
+         ("overlap identity (right)", OverlappingGCDTest checkOverlap_identity_right),
          ("isPrefixOf", LeftReductiveTest checkIsPrefixOf),
          ("stripSuffix", RightReductiveTest checkStripSuffix),
          ("isSuffixOf", RightReductiveTest checkIsSuffixOf),
@@ -472,7 +631,40 @@ tests = [("CommutativeMonoid", CommutativeTest checkCommutative),
          ("stripCommonSuffix 3", RightGCDTest checkStripCommonSuffix3),
          ("stripCommonSuffix 4", RightGCDTest checkStripCommonSuffix4),
          ("gcd", GCDTest checkGCD),
-         ("cancellative gcd", CancellativeGCDTest checkCancellativeGCD)
+         ("gcd uniqueness", GCDTest checkGCD_uniqueness),
+         ("gcd idempotence", GCDTest checkGCD_idempotence),
+         ("gcd identity (left)", GCDTest checkGCD_identity_left),
+         ("gcd identity (right)", GCDTest checkGCD_identity_right),
+         ("gcd commutativity", GCDTest checkGCD_commutativity),
+         ("gcd associativity", GCDTest checkGCD_associativity),
+         ("gcd distributivity (left)", DistributiveGCDTest checkGCD_distributivity_left),
+         ("gcd distributivity (right)", DistributiveGCDTest checkGCD_distributivity_right),
+         ("commonPrefix idempotence", LeftGCDTest checkCommonPrefix_idempotence),
+         ("commonPrefix identity (left)", LeftGCDTest checkCommonPrefix_identity_left),
+         ("commonPrefix identity (right)", LeftGCDTest checkCommonPrefix_identity_right),
+         ("commonPrefix commutativity", LeftGCDTest checkCommonPrefix_commutativity),
+         ("commonPrefix associativity", LeftGCDTest checkCommonPrefix_associativity),
+         ("commonPrefix distributivity", LeftDistributiveGCDTest checkCommonPrefix_distributivity),
+         ("commonSuffix idempotence", RightGCDTest checkCommonSuffix_idempotence),
+         ("commonSuffix identity (left)", RightGCDTest checkCommonSuffix_identity_left),
+         ("commonSuffix identity (right)", RightGCDTest checkCommonSuffix_identity_right),
+         ("commonSuffix commutativity", RightGCDTest checkCommonSuffix_commutativity),
+         ("commonSuffix associativity", RightGCDTest checkCommonSuffix_associativity),
+         ("commonSuffix distributivity", RightDistributiveGCDTest checkCommonSuffix_distributivity),
+         ("lcm reductivity (left)", LCMTest checkLCM_reductivity_left),
+         ("lcm reductivity (right)", LCMTest checkLCM_reductivity_right),
+         ("lcm uniqueness", LCMTest checkLCM_uniqueness),
+         ("lcm idempotence", LCMTest checkLCM_idempotence),
+         ("lcm identity (left)", LCMTest checkLCM_identity_left),
+         ("lcm identity (right)", LCMTest checkLCM_identity_right),
+         ("lcm commutativity", LCMTest checkLCM_commutativity),
+         ("lcm associativity", LCMTest checkLCM_associativity),
+         ("lcm absorption (gcd-lcm)", LCMTest checkLCM_absorption_gcd_lcm),
+         ("lcm absorption (lcm-gcd)", LCMTest checkLCM_absorption_lcm_gcd),
+         ("lcm distributivity (left)", DistributiveLCMTest checkLCM_distributivity_left),
+         ("lcm distributivity (right)", DistributiveLCMTest checkLCM_distributivity_right),
+         ("lcm distributivity (gcd-lcm)", DistributiveLCMTest checkLCM_distributivity_gcd_lcm),
+         ("lcm distributivity (lcm-gcd)", DistributiveLCMTest checkLCM_distributivity_lcm_gcd)
         ]
 
 checkCommutative (CommutativeMonoidInstance (e :: a)) = forAll (arbitrary :: Gen (a, a)) (\(a, b)-> a <> b == b <> a)
@@ -778,6 +970,24 @@ checkUnAppend (ReductiveMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen (a,
    where check (a, b) = maybe a (b <>) (a </> b) == a
                         && maybe a (<> b) (a </> b) == a
 
+checkOverlapLaw1 (OverlappingGCDMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen (a, a)) check
+   where check (a, b) = stripOverlap a b == (stripSuffixOverlap b a, overlap a b, stripPrefixOverlap a b)
+
+checkOverlapLaw2 (OverlappingGCDMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen (a, a)) check
+   where check (a, b) = stripSuffixOverlap b a <> overlap a b == a
+
+checkOverlapLaw3 (OverlappingGCDMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen (a, a)) check
+   where check (a, b) = overlap a b <> stripPrefixOverlap a b == b
+
+checkOverlap_idempotence (OverlappingGCDMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen a) $ \a -> overlap a a === a
+
+checkOverlap_identity_left (OverlappingGCDMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen a) $ \a -> overlap mempty a === mempty
+
+checkOverlap_identity_right (OverlappingGCDMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen a) $ \a -> overlap a mempty === mempty
+
 checkStripPrefixOverlap1 (OverlappingGCDMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen (a, a)) check
    where check (a, b) = o `isSuffixOf` b && b `isSuffixOf` (a <> o)
             where o = stripPrefixOverlap a b
@@ -863,11 +1073,181 @@ checkGCD (GCDMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen (a, a)) check
                         && isJust (b </> d)
             where d = gcd a b
 
-checkCancellativeGCD (CancellativeGCDMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen (a, a, a)) check
-   where check (a, b, c) = commonPrefix (a <> b) (a <> c) == a <> (commonPrefix b c)
-                           && commonSuffix (a <> c) (b <> c) == (commonSuffix a b) <> c
-                           && gcd (a <> b) (a <> c) == a <> gcd b c
-                           && gcd (a <> c) (b <> c) == gcd a b <> c
+checkGCD_uniqueness
+    (GCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a, a)) $
+        \(a, b, c) ->
+            all isJust [a </> c, b </> c, c </> gcd a b] === (gcd a b == c)
+
+checkGCD_idempotence
+    (GCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen a) $
+        \a -> gcd a a === a
+
+checkGCD_identity_left
+    (GCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen a) $
+        \a -> gcd mempty a === mempty
+
+checkGCD_identity_right
+    (GCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen a) $
+        \a -> gcd a mempty === mempty
+
+checkGCD_commutativity
+    (GCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a)) $
+        \a b -> gcd a b === gcd b a
+
+checkGCD_associativity
+    (GCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a, a)) $
+        \a b c -> gcd a (gcd b c) === gcd (gcd a b) c
+
+checkGCD_distributivity_left
+    (DistributiveGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a, a)) $
+        \(a, b, c) -> gcd (a <> b) (a <> c) == a <> gcd b c
+
+checkGCD_distributivity_right
+    (DistributiveGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a, a)) $
+        \(a, b, c) -> gcd (a <> c) (b <> c) == gcd a b <> c
+
+checkCommonPrefix_idempotence
+    (LeftGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen a) $
+        \a -> commonPrefix a a === a
+
+checkCommonPrefix_identity_left
+    (LeftGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen a) $
+        \a -> commonPrefix mempty a === mempty
+
+checkCommonPrefix_identity_right
+    (LeftGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen a) $
+        \a -> commonPrefix a mempty === mempty
+
+checkCommonPrefix_commutativity
+    (LeftGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a)) $
+        \a b -> commonPrefix a b === commonPrefix b a
+
+checkCommonPrefix_associativity
+    (LeftGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a, a)) $
+        \a b c ->
+            (commonPrefix a (commonPrefix b c)) ===
+            (commonPrefix (commonPrefix a b) c)
+
+checkCommonPrefix_distributivity
+    (LeftDistributiveGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a, a)) $
+        \(a, b, c) -> commonPrefix (a <> b) (a <> c) == a <> commonPrefix b c
+
+checkCommonSuffix_idempotence
+    (RightGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen a) $
+        \a -> commonSuffix a a === a
+
+checkCommonSuffix_identity_left
+    (RightGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen a) $
+        \a -> commonSuffix mempty a === mempty
+
+checkCommonSuffix_identity_right
+    (RightGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen a) $
+        \a -> commonSuffix a mempty === mempty
+
+checkCommonSuffix_commutativity
+    (RightGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a)) $
+        \a b -> commonSuffix a b === commonSuffix b a
+
+checkCommonSuffix_associativity
+    (RightGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a, a)) $
+        \a b c ->
+            (commonSuffix a (commonSuffix b c)) ===
+            (commonSuffix (commonSuffix a b) c)
+
+checkCommonSuffix_distributivity
+    (RightDistributiveGCDMonoidInstance (_ :: a)) =
+        forAll (arbitrary :: Gen (a, a, a)) $
+        \(a, b, c) -> commonSuffix (a <> c) (b <> c) == commonSuffix a b <> c
+
+checkLCM_reductivity_left (LCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a)) check
+  where
+    check a b = isJust (lcm a b </> a)
+
+checkLCM_reductivity_right (LCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a)) check
+  where
+    check a b = isJust (lcm a b </> b)
+
+checkLCM_uniqueness (LCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a, a)) check
+  where
+    check a b c =
+        all isJust [c </> a, c </> b, lcm a b </> c] === (lcm a b == c)
+
+checkLCM_idempotence (LCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen a) check
+  where
+    check a = lcm a a === a
+
+checkLCM_identity_left (LCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen a) check
+  where
+    check a = lcm mempty a === a
+
+checkLCM_identity_right (LCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen a) check
+  where
+    check a = lcm a mempty === a
+
+checkLCM_commutativity (LCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a)) check
+  where
+    check a b = lcm a b === lcm b a
+
+checkLCM_associativity (LCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a, a)) check
+  where
+    check a b c = lcm (lcm a b) c === lcm a (lcm b c)
+
+checkLCM_absorption_gcd_lcm (LCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a)) check
+  where
+    check a b = lcm a (gcd a b) === a
+
+checkLCM_absorption_lcm_gcd (LCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a)) check
+  where
+    check a b = gcd a (lcm a b) === a
+
+checkLCM_distributivity_left (DistributiveLCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a, a)) check
+  where
+    check a b c = lcm (a <> b) (a <> c) === a <> lcm b c
+
+checkLCM_distributivity_right (DistributiveLCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a, a)) check
+  where
+    check a b c = lcm (a <> c) (b <> c) === lcm a b <> c
+
+checkLCM_distributivity_gcd_lcm (DistributiveLCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a, a)) check
+  where
+    check a b c = lcm a (gcd b c) === gcd (lcm a b) (lcm a c)
+
+checkLCM_distributivity_lcm_gcd (DistributiveLCMMonoidInstance (_ :: a)) =
+    forAll (arbitrary :: Gen (a, a, a)) check
+  where
+    check a b c = gcd a (lcm b c) === lcm (gcd a b) (gcd a c)
 
 textualFactors :: TextualMonoid t => t -> [Either t Char]
 textualFactors = map characterize . factors
@@ -939,6 +1319,9 @@ instance (Arbitrary a, MonoidNull a, PositiveMonoid a) => Arbitrary (Concat a) w
 instance (Arbitrary a, FactorialMonoid a) => Arbitrary (Measured a) where
    arbitrary = fmap Measured.measure arbitrary
 
+instance (Arbitrary a, Monoid a) => Arbitrary (PrefixMemory.Shadowed a) where
+   arbitrary = fmap PrefixMemory.shadowed arbitrary
+
 instance (Arbitrary a, FactorialMonoid a) => Arbitrary (OffsetPositioned a) where
    arbitrary = fmap pure arbitrary
 
@@ -957,6 +1340,9 @@ instance CoArbitrary a => CoArbitrary (Concat a) where
 instance CoArbitrary a => CoArbitrary (Measured a) where
    coarbitrary = coarbitrary . Measured.extract
 
+instance CoArbitrary a => CoArbitrary (PrefixMemory.Shadowed a) where
+   coarbitrary = coarbitrary . PrefixMemory.content
+
 instance CoArbitrary a => CoArbitrary (OffsetPositioned a) where
    coarbitrary = coarbitrary . Positioned.extract
 
