@@ -1,5 +1,5 @@
 {-
-    Copyright 2014-2021 Mario Blazevic
+    Copyright 2014-2022 Mario Blazevic
 
     License: BSD3 (see BSD3-LICENSE.txt file)
 -}
@@ -51,7 +51,7 @@ class Positioned p where
 
 data OffsetPositioned m = OffsetPositioned{offset :: !Int,
                                            -- ^ the current offset
-                                           extractOffset :: m}
+                                           extractOffset :: m} deriving (Data, Typeable)
 
 data LinePositioned m = LinePositioned{fullOffset :: !Int,
                                        -- | the current line
@@ -502,13 +502,15 @@ instance (StableFactorial m, TextualMonoid m) => TextualMonoid (OffsetPositioned
    {-# INLINE concatMap #-}
    {-# INLINE foldl' #-}
    {-# INLINE foldr #-}
-   {-# INLINE spanMaybe' #-}
-   {-# INLINE span #-}
+   {-# INLINABLE spanMaybe #-}
+   {-# INLINABLE spanMaybe' #-}
+   {-# INLINABLE span #-}
    {-# INLINE foldl_' #-}
    {-# INLINE foldr_ #-}
    {-# INLINE any #-}
    {-# INLINE all #-}
-   {-# INLINE spanMaybe_' #-}
+   {-# INLINABLE spanMaybe_ #-}
+   {-# INLINABLE spanMaybe_' #-}
    {-# INLINE span_ #-}
    {-# INLINE break_ #-}
    {-# INLINE dropWhile_ #-}
@@ -674,8 +676,9 @@ instance (StableFactorial m, TextualMonoid m) => TextualMonoid (LinePositioned m
    {-# INLINE foldr_ #-}
    {-# INLINE any #-}
    {-# INLINE all #-}
-   {-# INLINE spanMaybe_' #-}
-   {-# INLINE span_ #-}
+   {-# INLINABLE spanMaybe_ #-}
+   {-# INLINABLE spanMaybe_' #-}
+   {-# INLINABLE span_ #-}
    {-# INLINE break_ #-}
    {-# INLINE dropWhile_ #-}
    {-# INLINE takeWhile_ #-}
